// ==========================================================================
// $Id$
// ==========================================================================
// (C)opyright: 2009
//
//   Ulm University
// 
// Creator: Hendrik Lensch, Holger Dammertz
// Email:   hendrik.lensch@uni-ulm.de, holger.dammertz@uni-ulm.de
// ==========================================================================
// $Log$
// ==========================================================================

#include <stdio.h>
#include <cutil.h> 
#include <vector_types.h>

using namespace std;
// using namespace base;
//using namespace image;

#define MAX_BLOCKS 256
#define MAX_THREADS 256


__global__ void dotProdKernel(float *_dst, const float* _a1, const
			      float* _a2, int _dim) {
  
  // calculate how many elements each thread needs to calculate 
  const unsigned int iter = _dim / (blockDim.x * gridDim.x); 
  int pos = blockIdx.x*MAX_THREADS + threadIdx.x; 
 
  // clear the output 

  float res = 0.; 

  for (int i = 0; i < iter; ++i) {

    //    _dst[ blockIdx.x*MAX_THREADS + threadIdx.x] += _a1[pos] * _a2[pos];
    res += _a1[pos] * _a2[pos];
    pos += blockDim.x * gridDim.x;  
  }
 
  // for the last iteration, check if the elements are still available
  if (pos < _dim) {
    //    _dst[ blockIdx.x*MAX_THREADS + threadIdx.x] += _a1[pos] * _a2[pos];
    res += _a1[pos] * _a2[pos];
  }

  if ( blockIdx.x*MAX_THREADS + threadIdx.x < _dim) 
    _dst[ blockIdx.x*MAX_THREADS + threadIdx.x] = res; 
  
   
}

/** This kernel first uploads all elements of _in to the shared
    memory. 

    The kernel computes the sum of all elements corresponding to all
    threads wihtin one block (in parallel). In each step the
    number of active threads per block is divided by two. 

    At the end, the sum of this block is stored at _dst[blockIdx.x]

    You can assume that the maximum number of threads per block is 256
   
*/     
__global__ void reduceSumKernel(int n, float *_dst, const float *_in) {
  
  extern __shared__ int sdata[]; 

  // upload to shared memory
  unsigned int bs = blockDim.x; // blocksize 
  unsigned int bid = blockIdx.x * blockDim.x;  // offset for this block
  unsigned int tid = threadIdx.x; 

  unsigned int pos = bid + tid; 


  //////////////////////////////////////////
  //!!!! missing 
  //////////////////////////////////////////


  // store the final result of this block 
  if (tid == 0) _dst[blockIdx.x] = (float)sdata[0]; 

}



 

  
/* This program sets up two large arrays of size dim and computes the
   dot product of both arrays. 

   Most of the code of previous exercises is reused. 
   Mode 0 of the program computes the final dot product as before. 

   Mode 1: After computing the dot product and storing the result for all 
   MAX_BLOCKS * MAX_THREAD threads, this time, the reduction of the sum
   is to be computed on the GPU. 

   Write a reduction sum kernel which is called log(n) times. 
   The number of total threads will be divided by nThreads(iter-1)
   in each iteration. 
   
   Inside the kernel, the problem will be reduced by a factor of 2 in
   each step. 

*/ 
   

int main(int argc, char* argv[]) {

  // parse command line 
  int acount = 1; 

  if (argc < 3) { 
    printf("usage: testDotProductStreams <dim> <copy mode [0,1,2]>"); 
    exit(1) ;
  }

  // number of elements in both vectors 
  int dim = atoi(argv[acount++]); 

  // mode of the memory upload
  int mode = atoi( argv[acount++] ); 


  printf("dim: %d\n", dim); 

  if (dim < 128) {
      printf("does not work for dim < 128 yet\n"); 
      //      exit(1); 
  }


  float* cpuArray1 = new float[dim]; 

  // extra array that points to non-pagelocked array for faster transfer
  float* cpuArray2Pagelocked; 

  // create the pagelocked array of dimension dim
  cudaMallocHost( (void**)&cpuArray2Pagelocked, dim * sizeof(float)); 

  // initialize the two arrays 
  for (int i = 0; i < dim; ++i) {
    
#ifdef RTEST 
    cpuArray1[i] = drand48(); 
    cpuArray2[i] = drand48();
#else 
     cpuArray1[i] = 1.0;
     cpuArray2Pagelocked[i] = 1.; // i % 10; 
#endif 
  }

  // now the gpu stuff 
  float* gpuArray1; 
  float* gpuArray2; 

  // you might need two result arrays for the kernels of the two streams
  float* gpuResult[2]; 
  

  
  float* h; 

  // setup cuda environment

  //  CUT_DEVICE_INIT();	

  // allocate two gpuArray 1  and gpuArray 2 and gpuResult array on GPU
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuArray1, dim * sizeof(float) ) );   
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuArray2, dim * sizeof(float) ) );      
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuResult[0], MAX_BLOCKS *
  MAX_THREADS * sizeof(float) ) );   
  CUDA_SAFE_CALL( cudaMemset( gpuResult[0], 0, MAX_BLOCKS * MAX_THREADS *
			      sizeof(float)) ); 

  // result[1] is only used to store the result of the reduction kernel
  // its maximum size is MAX_BLOCKS
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuResult[1], MAX_BLOCKS * sizeof(float) ) );   
  

  // copy array 1 once to the device
  CUDA_SAFE_CALL( cudaMemcpy( gpuArray1, cpuArray1, dim * sizeof(float),  cudaMemcpyHostToDevice) );
  
  
  // allocate an array to download the results of all threads 
  // for downloading the reduction result, only the first entry will be used
  cudaMallocHost( (void**)&h, MAX_THREADS*MAX_BLOCKS * sizeof(float)); 

  double finalDotProduct = 0.; 

  for (int iter = 0; iter < 1000; ++iter) {

    finalDotProduct = 0.; 

    // a simplistic way of splitting the problem into threads
    dim3 blockGrid( MAX_BLOCKS ); 
    dim3 threadBlock(MAX_THREADS); 


    int nThreads = dim; 
    if (nThreads > MAX_BLOCKS * MAX_THREADS) {
      nThreads = MAX_BLOCKS * MAX_THREADS;
    }
    
    CUDA_SAFE_CALL( cudaMemcpy( gpuArray2, cpuArray2Pagelocked, dim * sizeof(float),  cudaMemcpyHostToDevice) );


    switch (mode) {

    case 0: 


      //      printf( "cpu reduction: "); 
      
      // update Array 2

      // call the dot kernel 
      dotProdKernel<<< blockGrid, threadBlock>>>(gpuResult[0], gpuArray1, gpuArray2,dim); 

      // check that everything went soothly
      CUT_CHECK_ERROR("dot product execution failed\n");
      CUDA_SAFE_CALL( cudaThreadSynchronize() );	

      // download and combine the results of multiple threads  
      CUDA_SAFE_CALL( cudaMemcpy( h, gpuResult[0], nThreads * sizeof(float),  cudaMemcpyDeviceToHost) );
      
      finalDotProduct = 0.; 

      // accumulate the final result on the host
      for (int i = 0; i < nThreads; ++i) {
	finalDotProduct += h[i]; 
      }

      break; 
 
    case 1: 

      //      printf( "gpu reduction: "); 
     
      
      // compute the dot product into result1
      dotProdKernel<<< blockGrid, threadBlock>>>(gpuResult[1], gpuArray1, gpuArray2,dim); 
      CUT_CHECK_ERROR("dot product execution failed\n");
      CUDA_SAFE_CALL( cudaThreadSynchronize() );


      // sum up all the elements within each block
      // i.e. run the reductions for all blocks
      reduceSumKernel<<< blockGrid, threadBlock, MAX_THREADS*sizeof(float) >>>( MAX_THREADS*MAX_THREADS, gpuResult[0], gpuResult[1] ); 
      CUT_CHECK_ERROR("reduction call 1 failed\n");
      CUDA_SAFE_CALL( cudaThreadSynchronize() );
      
      // call the reduction kernel a second time with a single block 
      // as we have 256 blocks and 256 threads this reduces the problem to a single number  
      dim3 blockGrid2( 1, 1 ); 
      reduceSumKernel<<< blockGrid2, threadBlock, MAX_THREADS*sizeof(float) >>>( MAX_THREADS, gpuResult[1], gpuResult[0] ); 

      CUT_CHECK_ERROR("sum Reduction call 2 failed\n");
      CUDA_SAFE_CALL( cudaThreadSynchronize() );	
      
      // download the final number
      CUDA_SAFE_CALL( cudaMemcpy( h, gpuResult[1], 1*sizeof(float),  cudaMemcpyDeviceToHost) );

      finalDotProduct = h[0]; 
      
      break; 


    } // end switch
      


       
  }

  printf( "%f\n", finalDotProduct); 
  
  // cleanup GPU memory
  
  CUDA_SAFE_CALL( cudaFree(gpuResult[1]) ); 
  CUDA_SAFE_CALL( cudaFree(gpuResult[0]) ); 
  CUDA_SAFE_CALL( cudaFree(gpuArray2) ); 
  CUDA_SAFE_CALL( cudaFree(gpuArray1) ); 

  // free page locked memory
  cudaFreeHost( cpuArray2Pagelocked ); 
  cudaFreeHost( h ); 

  delete[] cpuArray1; 

  printf("done\n"); 

}
